#include "trigrp.h"
#include "ali.h"
#include "outstream.h"

TriGrp::TriGrp () {
  nSeq_ = Ali::getNSeq ();
  itemArr_.set (nSeq_ * 2, nSeq_ * 2);
  genDisMat_ = Ali::getGenDis ();
}

void TriGrp::clean () {
  for (int seqNo_ = 0; seqNo_ < nSeq_; seqNo_ ++)
    itemArr_[seqNo_].clean ();

  nCurTree_ = nSeq_;
}

//-------------------------------------------------------------
//compute the observed distance among three sequences
double TriGrp::cmpOutGrpDis (int outGrpNdNo, int lNdNo, int rNdNo) {
  double dis12_ = genDisMat_[outGrpNdNo][lNdNo];
  double dis13_ = genDisMat_[outGrpNdNo][rNdNo];
  double dis23_ = genDisMat_[lNdNo][rNdNo];
  double hafSumDis_ = (dis12_ + dis13_ + dis23_) / 2.0;
  return hafSumDis_  - dis23_;
}

//-------------------------------------------------------------
//compute the observed distance among the outgrpnd and two representive colections
double TriGrp::cmpOutGrpDis (int outGrpNdNo, Vec<int> &lRepLeaArr, Vec<int> &rRepLeaArr) {
  int nLRepLea_ = lRepLeaArr.getSize ();
  int nRRepLea_ = rRepLeaArr.getSize ();
  double outGrpDis_ = 0.0;
//  double totalWei_ = 0.0;

  for (int lCount_ = 0; lCount_ < nLRepLea_; lCount_ ++) 
    for (int rCount_ = 0; rCount_ < nRRepLea_; rCount_ ++) {
       int lRepLeaNo_ = lRepLeaArr[lCount_];
       int rRepLeaNo_ = rRepLeaArr[rCount_];
       double dis12_ = genDisMat_[outGrpNdNo][lRepLeaNo_];
       double dis13_ = genDisMat_[outGrpNdNo][rRepLeaNo_];
       double dis23_ = genDisMat_[lRepLeaNo_][rRepLeaNo_];

       outGrpDis_ += (dis12_ + dis13_ - dis23_) / 2.0;
    }

  return  outGrpDis_ /  (nLRepLea_ * nRRepLea_);
}    

//====================================================================================
void TriGrp::init (int outGrpNdNo) {
  outGrpNdNo_ = outGrpNdNo;
  for (int seqNo_ = 0; seqNo_ < nSeq_; seqNo_ ++)
    if (seqNo_ != outGrpNdNo_) 
      itemArr_[seqNo_].init (seqNo_);

  for (int treeNo1_ = 0; treeNo1_ < nSeq_; treeNo1_ ++)
    if (treeNo1_ != outGrpNdNo_) 
      for (int treeNo2_ = treeNo1_ + 1; treeNo2_ < nSeq_; treeNo2_ ++)
        if (treeNo2_ != outGrpNdNo_) {
          double outGrpDis_ = cmpOutGrpDis (outGrpNdNo, treeNo1_, treeNo2_);

          if (itemArr_[treeNo1_].getNeaSibDis () < outGrpDis_) 
             itemArr_[treeNo1_].setNeaSib (treeNo2_, outGrpDis_);

          if (itemArr_[treeNo2_].getNeaSibDis () < outGrpDis_) 
             itemArr_[treeNo2_].setNeaSib (treeNo1_, outGrpDis_);
        }
}



//====================================================================================
void TriGrp::findPairTree (int &rootNo1, int &rootNo2) {
  double maxDis_ = -INFINITIVE;
  rootNo1 = -1;
  rootNo2 = -1;

  for (int treeNo_ = 0; treeNo_ < nCurTree_; treeNo_ ++)
    if ( (itemArr_[treeNo_].isActive () == 1) &&
      (itemArr_[treeNo_].getNeaSibDis () > maxDis_) ) {
      maxDis_ = itemArr_[treeNo_].getNeaSibDis ();
      rootNo1 = treeNo_;
      rootNo2 = itemArr_[treeNo_].getNeaSibNo ();
    }
}


void TriGrp::updateSibInf (RootedTree &newTree, int oldLRootNo, int oldRRootNo) {
  int newTreeNo_ = newTree.getRoot ();

  for (int treeNo_ = 0; treeNo_ < nCurTree_; treeNo_ ++)
    if ( (itemArr_[treeNo_].isActive () == 1) && treeNo_ != newTreeNo_) {
      double outGrpDis_ = cmpOutGrpDis (outGrpNdNo_, *newTree.getRepLea (), *itemArr_[treeNo_].getRepLea ());



      if (newTree.getNeaSibDis () < outGrpDis_) 
        newTree.setNeaSib (treeNo_, outGrpDis_);

      if (itemArr_[treeNo_].getNeaSibDis () < outGrpDis_ || itemArr_[treeNo_].getNeaSibNo () == oldLRootNo || itemArr_[treeNo_].getNeaSibNo () == oldRRootNo)
        itemArr_[treeNo_].setNeaSib (newTreeNo_, outGrpDis_);
    }
}

void TriGrp::constructOneTree (int outGrpNdNo) {  
  clean ();
  init (outGrpNdNo);
  

  for (int count_ = 0; count_ < nSeq_ - 2; count_ ++) {
    int rootNo1_, rootNo2_;
    findPairTree (rootNo1_, rootNo2_);
//    std::cout << nCurTree_ << " / " << rootNo1_ << " / " << rootNo2_ << endl;
    int newTreeNo_ = nCurTree_;
    
    itemArr_[newTreeNo_].mergeTwoTree (outGrpNdNo_, newTreeNo_, nRepLea_,  &itemArr_[rootNo1_], &itemArr_[rootNo2_]); 
    itemArr_[newTreeNo_].correct (nRepLea_);
    
    itemArr_[newTreeNo_].turnOn ();
    updateSibInf (itemArr_[newTreeNo_], rootNo1_, rootNo2_);

    nCurTree_ ++;
  }
}


//===============================================================================
//two follow functions serve to find the best moderate sequence and think of it as the referred sequence
double TriGrp::cmpFaDis (int outGrpNdNo) {
  double faDis_ = 0.0;
  for (int seqNo_ = 0; seqNo_ < nSeq_; seqNo_ ++)
    faDis_ = Utl::getMax (faDis_, genDisMat_[seqNo_][outGrpNdNo]);
  return faDis_;
}

int TriGrp::findBestOutGrpNd () {
  int bestOutGrpNdNo_ = -1;
  double minDis_ = INFINITIVE;
  for (int seqNo_ = 0; seqNo_ < nSeq_; seqNo_ ++) {
    double faDis_ = cmpFaDis (seqNo_);
    if (faDis_ < minDis_) {
      minDis_ = faDis_;
      bestOutGrpNdNo_ = seqNo_;
    }
  }

  return bestOutGrpNdNo_;
}

//============================================================================== 
void TriGrp::createTree (Vec<char> &tree, int nRepLea) { 
  nSeq_ = Ali::getNSeq ();
  nRepLea_ = nRepLea;
  
  int bestOutGrpNdNo_  = findBestOutGrpNd ();
  constructOneTree (bestOutGrpNdNo_);
//  std::cout << bestOutGrpNdNo_ << endl;
  int lastTreeNo_ = nCurTree_ - 1;
  itemArr_[lastTreeNo_].createNwTree (bestOutGrpNdNo_, nRepLea, tree);
}

TriGrp::~TriGrp () {
}

